#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .datasets.audiocaps import AudioCaps
from .datasets.clotho import Clotho
from .datasets.macs import MACS


__author__ = "Etienne Labbé (Labbeti)"
__maintainer__ = "Etienne Labbé (Labbeti)"
__name__ = "aac_datasets"
__status__ = "Development"
__version__ = "0.1.1"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys

from argparse import ArgumentParser, Namespace
from typing import Dict

import yaml

from aac_datasets.datasets.audiocaps import AudioCaps
from aac_datasets.datasets.clotho import Clotho
from aac_datasets.datasets.macs import MACS


logger = logging.getLogger(__name__)


def get_main_check_args() -> Namespace:
    parser = ArgumentParser(description="Check datasets in specified directory.")

    parser.add_argument(
        "--root",
        type=str,
        default=".",
        help="The path to the parent directory of the datasets.",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        default=1,
        help="Verbose level of the script. 0 means silent mode, 1 is default mode and 2 add additional debugging outputs.",
    )

    args = parser.parse_args()
    return args


def check_datasets(root: str, verbose: int = 0) -> Dict[str, Dict[str, int]]:
    datasets_lens = {"audiocaps": {}, "clotho": {}, "macs": {}}

    if verbose >= 1:
        logger.info(f"Searching for audiocaps in root='{root}'...")

    for subset in AudioCaps.SUBSETS:
        try:
            dataset = AudioCaps(root, subset, verbose=0)
            _ = dataset[0]
            datasets_lens["audiocaps"][subset] = len(dataset)
        except RuntimeError:
            if verbose >= 2:
                logger.info(f"Cannot find audiocaps_{subset}.")

    if verbose >= 1:
        logger.info(f"Searching for clotho in root='{root}'...")

    for subset in Clotho.SUBSETS:
        try:
            dataset = Clotho(root, subset, verbose=0)
            _ = dataset[0]
            datasets_lens["clotho"][subset] = len(dataset)
        except RuntimeError:
            if verbose >= 2:
                logger.info(f"Cannot find clotho_{subset}.")

    if verbose >= 1:
        logger.info(f"Searching for macs in root='{root}'...")

    try:
        dataset = MACS(root, verbose=0)
        _ = dataset[0]
        datasets_lens["macs"]["full"] = len(dataset)
    except RuntimeError:
        if verbose >= 2:
            logger.info("Cannot find macs.")

    datasets_lens = {
        dataset_name: subsets_lens
        for dataset_name, subsets_lens in datasets_lens.items()
        if len(subsets_lens) > 0
    }
    return datasets_lens


def main_check() -> None:
    format_ = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(format_))
    logger = logging.getLogger("aac_datasets")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    args = get_main_check_args()

    if args.verbose >= 2:
        print(yaml.dump({"Arguments": args.__dict__}, sort_keys=False))

    datasets_lens = check_datasets(args.root, args.verbose)

    if args.verbose >= 1:
        print(f"Found {len(datasets_lens)} dataset(s) in root='{args.root}'.")
        if len(datasets_lens) > 0:
            print(yaml.dump(datasets_lens, sort_keys=False))


if __name__ == "__main__":
    main_check()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import List, Tuple

from torch import Tensor
from torch.nn.utils.rnn import pad_sequence


class BasicCollate:
    def __init__(self, pad_value: float = 0.0) -> None:
        super().__init__()
        self.pad_value = pad_value

    def __call__(self, batch: List[Tuple]) -> Tuple:
        audio_batch = [item[0] for item in batch]
        captions_batch = [item[1] for item in batch]

        if len(audio_batch) == 0:
            raise ValueError("Cannot collate an empty list of audio.")

        are_tensors = [isinstance(audio, Tensor) for audio in audio_batch]
        if not all(are_tensors):
            raise TypeError(
                f"Invalid audio type in {self.__class__.__name__}. (found {are_tensors=})"
            )

        are_paddable = [
            audio.ndim > 0 and audio.shape[1:] == audio_batch[0].shape[1:]
            for audio in audio_batch
        ]
        if not all(are_paddable):
            raise TypeError(
                f"Invalid audio shapes in {self.__class__.__name__}. (found {are_paddable=})"
            )

        audio_batch = pad_sequence(
            audio_batch,
            batch_first=True,
            padding_value=self.pad_value,
        )
        return audio_batch, captions_batch



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import tqdm
import os
import os.path as osp
import shutil
import zipfile

from dataclasses import asdict, astuple, dataclass, field, fields
from functools import cached_property
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torchaudio
import yaml

from torch import nn, Tensor
from torch.utils.data.dataset import Dataset
from torchaudio.backend.common import AudioMetaData
from torchaudio.datasets.utils import download_url


logger = logging.getLogger(__name__)


@dataclass
class MACSItem:
    audio: Tensor = torch.empty((0,))
    captions: List[str] = field(default_factory=list)
    tags: List[List[str]] = field(default_factory=list)
    annotators_ids: List[str] = field(default_factory=list)
    fname: Optional[str] = None
    index: int = -1
    dataset: str = "macs"
    sr: int = -1


class MACS(Dataset):
    """
    Unofficial MACS pytorch dataset.

    Dataset folder tree:

    ```
    {root}
    └── MACS
        ├── audio
        │    └── (3930 files, ~13GB)
        ├── LICENCE.txt
        ├── MACS.yaml
        └── MACS_competence.csv
    ```
    """

    AUDIO_MAX_LENGTH = 10.0  # in seconds
    AUDIO_N_CHANNELS = 2
    CLEAN_ARCHIVES: bool = False
    FORCE_PREPARE_DATA: bool = False
    ITEM_TYPES = ("tuple", "dict", "dataclass")
    MAX_CAPTIONS_PER_AUDIO = {"full": 5}
    MIN_CAPTIONS_PER_AUDIO = {"full": 2}
    SAMPLE_RATE = 48000  # in Hz
    SUBSETS = ("full",)

    def __init__(
        self,
        root: str = ".",
        download: bool = False,
        transforms: Optional[Dict[str, Optional[nn.Module]]] = None,
        unfold: bool = False,
        item_type: str = "tuple",
        multihot_tags: bool = False,
        verbose: int = 0,
    ) -> None:
        """
        :param root: The parent of the dataset root directory.
            The data will be stored in the 'MACS' subdirectory.
            defaults to ".".
        :param download: Download the dataset if download=True and if the dataset is not already downloaded.
            defaults to False.
        :param transforms: The transform to apply to values.
            Keys can be 'audio', 'captions' or 'tags'.
            defaults to None.
        :param unfold: If True, map captions to audio instead of audio to caption.
            defaults to True.
        :param item_type: The type of the value returned by __getitem__.
            Can be 'tuple', 'dict' or 'dataclass'.
            defaults to 'tuple'.
        :param multihot_tags: If True, tags are loaded as multihot tensors of 10 classes.
            Otherwise the tags names are returns as a list of str.
            defaults to False.
        :param verbose: Verbose level to use. Can be 0 or 1.
            defaults to 0.
        """
        if item_type not in self.ITEM_TYPES:
            raise ValueError(
                f"Invalid argument {item_type=} for MACS. (expected one of {self.ITEM_TYPES})"
            )

        if transforms is None:
            transforms = {}

        accepted_keys = [field_.name for field_ in fields(MACSItem)]
        for key in transforms.keys():
            if key not in accepted_keys:
                raise ValueError(
                    f"Invalid transform {key=}. (expected one of {accepted_keys})"
                )

        super().__init__()
        self._root = root
        self._download = download
        self._transforms = transforms
        self._unfold = unfold
        self._item_type = item_type
        self._multihot_tags = multihot_tags
        self._verbose = verbose

        self._all_infos = []
        self._tag_to_idx = {}
        self._idx_to_tag = {}

        if self._download:
            self._prepare_data()
        self._load_data()

    def __len__(self) -> int:
        return len(self._all_infos)

    def __getitem__(self, index: int) -> Union[Tuple, Dict, MACSItem]:
        kwargs = {field.name: self.get(field.name, index) for field in fields(MACSItem)}
        item = MACSItem(**kwargs)

        if self._item_type == "tuple":
            return astuple(item)
        elif self._item_type == "dict":
            return asdict(item)
        elif self._item_type == "dataclass":
            return item
        else:
            raise ValueError(
                f"Invalid item_type={self._item_type} for MACS. (expected one of {self.ITEM_TYPES})"
            )

    def get_raw(self, name: str, index: int) -> Any:
        """Read the processed raw data. (without transform)

        :param name: The name of the value.
            Can be 'audio', 'fpath', 'fname', 'captions', 'tags', 'annotators_ids'.
        :param index: The index of the value in range [0, len(dataset)[.
        """
        if not (0 <= index < len(self)):
            raise IndexError(
                f"Invalid argument {index=} for {self} (expected in range [0, {len(self)}-1])"
            )

        if name == "audio":
            fpath = self.get("fpath", index)
            value, sr = torchaudio.load(fpath)  # type: ignore

            if value.nelement() == 0:
                raise RuntimeError(f'Invalid audio number of elements in "{fpath}".')
            if sr != self.SAMPLE_RATE:
                raise RuntimeError(f'Invalid sample rate in "{fpath}".')

        elif name == "fpath":
            fname = self.get("fname", index)
            value = osp.join(self._dpath_audio, fname)

        elif name == "dataset":
            value = "macs"

        elif name == "index":
            value = index

        elif name == "sr":
            fpath = self.get("fpath", index)
            audio_info: AudioMetaData = torchaudio.info(fpath)  # type: ignore
            return audio_info.sample_rate

        elif name in self._all_infos[index].keys():
            value = self._all_infos[index][name]

        else:
            raise ValueError(f"Invalid value {name=} at {index=}.")
        return value

    def get(self, name: str, index: int) -> Any:
        """Read the processed raw data. (with transform)

        :param name: The name of the value.
                Can be 'audio', 'fpath', 'fname', 'captions', 'tags', 'annotators_ids'.
        :param index: The index of the value in range [0, len(dataset)[.
        """
        value = self.get_raw(name, index)
        transform = self._transforms.get(name, None)
        if transform is not None:
            value = transform(value)
        return value

    def tag_to_idx(self, tag: str) -> int:
        return self._tag_to_idx[tag]

    def idx_to_tag(self, idx: int) -> str:
        return self._idx_to_tag[idx]

    def _is_prepared(self) -> bool:
        if not osp.isdir(self._dpath_audio):
            return False
        captions_fpath = osp.join(self._dpath_data, MACS_FILES["captions"]["fname"])
        if not osp.isfile(captions_fpath):
            return False

        with open(captions_fpath, "r") as file:
            data = yaml.safe_load(file)
        data = data["files"]
        fnames = os.listdir(self._dpath_audio)
        return len(data) == len(fnames)

    def _prepare_data(self) -> None:
        if not osp.isdir(self._root):
            raise RuntimeError(f"Cannot find root directory '{self._root}'.")

        os.makedirs(self._dpath_audio, exist_ok=True)
        os.makedirs(self._dpath_archives, exist_ok=True)

        for file_info in MACS_FILES.values():
            dpath = self._dpath_data
            fname = file_info["fname"]
            fpath = osp.join(dpath, fname)
            if not osp.isfile(fpath) or self.FORCE_PREPARE_DATA:
                if self._verbose >= 1:
                    logger.info(f"Downloading captions file '{fname}'...")

                url = MACS_FILES["captions"]["url"]
                hash_value = MACS_FILES["captions"]["hash"]
                download_url(
                    url,
                    dpath,
                    fname,
                    hash_value,
                    "md5",
                    progress_bar=self._verbose >= 1,
                    resume=True,
                )

        captions_fpath = osp.join(self._dpath_data, MACS_FILES["captions"]["fname"])
        with open(captions_fpath, "r") as file:
            captions_data = yaml.safe_load(file)
        captions_data = captions_data["files"]

        for i, infos in enumerate(MACS_ARCHIVES_FILES.values()):
            dpath = self._dpath_archives
            zip_fname = infos["fname"]
            zip_fpath = osp.join(dpath, zip_fname)

            if not osp.isfile(zip_fpath) or self.FORCE_PREPARE_DATA:
                if self._verbose >= 1:
                    logger.info(
                        f"Downloading audio zip file '{zip_fpath}'... ({i+1}/{len(MACS_ARCHIVES_FILES)})"
                    )

                url = infos["url"]
                hash_value = infos["hash"]
                download_url(
                    url,
                    dpath,
                    zip_fname,
                    hash_value,
                    "md5",
                    progress_bar=self._verbose >= 1,
                )

        macs_fnames = dict.fromkeys(data["filename"] for data in captions_data)
        for i, infos in enumerate(MACS_ARCHIVES_FILES.values()):
            zip_fname = infos["fname"]
            zip_fpath = osp.join(self._dpath_archives, zip_fname)

            with zipfile.ZipFile(zip_fpath, "r") as file:
                members_to_extract = [
                    member
                    for member in file.namelist()
                    if osp.basename(member) in macs_fnames
                    and not osp.isfile(
                        osp.join(self._dpath_audio, osp.basename(member))
                    )
                ]

                if self._verbose >= 1:
                    logger.info(
                        f"Extracting {len(members_to_extract)}/{len(file.namelist())} audio files from ZIP file '{zip_fname}'... ({i+1}/{len(MACS_ARCHIVES_FILES)})"
                    )

                if len(members_to_extract) > 0:
                    file.extractall(self._dpath_archives, members_to_extract)
                    for member in members_to_extract:
                        extracted_fpath = osp.join(self._dpath_archives, member)
                        target_fpath = osp.join(self._dpath_audio, osp.basename(member))
                        shutil.move(extracted_fpath, target_fpath)

        if self.CLEAN_ARCHIVES:
            if self._verbose >= 1:
                logger.info(f"Removing archives files in {self._dpath_archives}...")
            shutil.rmtree(self._dpath_archives, ignore_errors=True)

        audio_fnames = [
            name for name in os.listdir(self._dpath_audio) if name.endswith(".wav")
        ]
        audio_fpaths = [osp.join(self._dpath_audio, name) for name in audio_fnames]
        assert len(audio_fpaths) == len(macs_fnames)

        if self._verbose >= 1:
            logger.info(
                f"{len(audio_fpaths)} audio files has been prepared for MACS dataset."
            )

    def _load_data(self) -> None:
        if not self._is_prepared():
            raise RuntimeError(
                f"Cannot load data: macs is not prepared in data root={self._root}. Please use download=True in dataset constructor."
            )

        captions_fpath = osp.join(self._dpath_data, MACS_FILES["captions"]["fname"])
        if self._verbose >= 1:
            logger.debug(f'Loading captions file "{captions_fpath}"...')

        with open(captions_fpath, "r") as file:
            data = yaml.safe_load(file)

        self._all_infos = [
            {
                "fname": item["filename"],
                "captions": [subitem["sentence"] for subitem in item["annotations"]],
                "tags": [subitem["tags"] for subitem in item["annotations"]],
                "annotators_ids": [
                    subitem["annotator_id"] for subitem in item["annotations"]
                ],
            }
            for item in data["files"]
        ]

        all_tags = [item["tags"] for item in self._all_infos]
        all_tags_flat = (
            tag
            for tags in all_tags
            for tags_by_annotator in tags
            for tag in tags_by_annotator
        )
        all_tags_uniq = dict.fromkeys(all_tags_flat)
        self._tag_to_idx = {tag: i for i, tag in enumerate(all_tags_uniq)}
        self._idx_to_tag = {i: tag for i, tag in enumerate(all_tags_uniq)}

        if self._multihot_tags:
            n_items = len(all_tags)
            max_annotators = max(map(len, all_tags))
            max_tag_by_annotator = max(
                map(
                    len,
                    (
                        tags_by_annotator
                        for tags in all_tags
                        for tags_by_annotator in tags
                    ),
                )
            )
            n_tags_uniq = len(all_tags_uniq)

            multihots_tags_shape = (
                n_items,
                max_annotators,
                max_tag_by_annotator,
                n_tags_uniq,
            )
            multihots_tags = torch.zeros(multihots_tags_shape, dtype=torch.bool)

            for i, tags in enumerate(tqdm.tqdm(all_tags, disable=self._verbose == 0)):
                for j, tags_by_annotator in enumerate(tags):
                    for k, tag in enumerate(tags_by_annotator):
                        idx = self._tag_to_idx[tag]
                        multihots_tags[i, j, k, idx] = True

            for i, multihots in enumerate(multihots_tags):
                self._all_infos[i]["tags"] = multihots

        if self._unfold and self.MIN_CAPTIONS_PER_AUDIO["full"] > 1:
            data_info_unfolded = []
            for infos in self._all_infos:
                captions = infos["captions"]
                for caption in captions:
                    new_infos = dict(infos)
                    new_infos["captions"] = (caption,)
                    data_info_unfolded.append(new_infos)

            self._all_infos = data_info_unfolded

        if self._verbose >= 1:
            logger.info(f"{self.__class__.__name__} has been loaded. (len={len(self)})")

    @cached_property
    def _dpath_data(self) -> str:
        return osp.join(self._root, "MACS")

    @cached_property
    def _dpath_audio(self) -> str:
        return osp.join(self._dpath_data, "audio")

    @cached_property
    def _dpath_archives(self) -> str:
        return osp.join(self._dpath_data, "archives")

    def __repr__(self) -> str:
        return "MACS()"


MACS_FILES = {
    "licence": {
        "fname": "LICENSE.txt",
        "url": "https://zenodo.org/record/5114771/files/LICENSE.txt?download=1",
        "hash": "d3086f4517cccc32c1bb3a081b07cfa1",
    },
    "captions": {
        "fname": "MACS.yaml",
        "url": "https://zenodo.org/record/5114771/files/MACS.yaml?download=1",
        "hash": "23fcb2ebd0b109094034ef9e87972256",
    },
    "annotators_competences": {
        "fname": "MACS_competence.csv",
        "url": "https://zenodo.org/record/5114771/files/MACS_competence.csv?download=1",
        "hash": "4dfe9f951f0af9f29cb7952ec030370a",
    },
}

TAU_URBAN_ACOUSTIC_DEV_FILES = {
    "audio.1": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.1.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.1.zip?download=1",
        "hash": "aca4ebfd9ed03d5f747d6ba8c24bc728",
    },
    "audio.10": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.10.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.10.zip?download=1",
        "hash": "0ffbf60006da520cc761fb74c878b98b",
    },
    "audio.11": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.11.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.11.zip?download=1",
        "hash": "599055d93b4c11057c29be2df54538d4",
    },
    "audio.12": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.12.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.12.zip?download=1",
        "hash": "98b8d162ff3665695c4c910e6c372cc8",
    },
    "audio.13": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.13.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.13.zip?download=1",
        "hash": "a356c08b1a5a21d433eba37ef87587f4",
    },
    "audio.14": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.14.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.14.zip?download=1",
        "hash": "f8969771e7faf7dd471d1cf78b0cf011",
    },
    "audio.15": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.15.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.15.zip?download=1",
        "hash": "4758c4b0fb7484faa632266e78850820",
    },
    "audio.16": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.16.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.16.zip?download=1",
        "hash": "a18acad9ede8ea76574216feb887f0bc",
    },
    "audio.17": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.17.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.17.zip?download=1",
        "hash": "1af7703484632f340da5c33662dc9632",
    },
    "audio.18": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.18.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.18.zip?download=1",
        "hash": "b67402bf3e08f4da394a7c18756c0fd2",
    },
    "audio.19": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.19.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.19.zip?download=1",
        "hash": "035db315f19106eb848b6f9b32bcc47c",
    },
    "audio.2": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.2.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.2.zip?download=1",
        "hash": "c4f170408ce77c8c70c532bf268d7be0",
    },
    "audio.20": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.20.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.20.zip?download=1",
        "hash": "9cb28c74911bf8a3eadcf53f50a5b5d6",
    },
    "audio.21": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.21.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.21.zip?download=1",
        "hash": "0e44ed85c88ec036a9725b4dd1dfaea0",
    },
    "audio.3": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.3.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.3.zip?download=1",
        "hash": "c7214a07211f10f3250290d05e72c37e",
    },
    "audio.4": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.4.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.4.zip?download=1",
        "hash": "a6a62110f6699cf4432072acb1dffda6",
    },
    "audio.5": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.5.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.5.zip?download=1",
        "hash": "091a0b6d3c84b8e60e46940aa7d4a8a0",
    },
    "audio.6": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.6.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.6.zip?download=1",
        "hash": "114f4ca13e074391b98a1cfd8140de65",
    },
    "audio.7": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.7.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.7.zip?download=1",
        "hash": "5951dd2968f7a514e2afbe279c4f060d",
    },
    "audio.8": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.8.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.8.zip?download=1",
        "hash": "b0b63dc95b327e1509857c8d8a663cc3",
    },
    "audio.9": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.audio.9.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.audio.9.zip?download=1",
        "hash": "3c32a693a6b111ffb957be3c1dd22e9b",
    },
    "doc": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.doc.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.doc.zip?download=1",
        "hash": "1f6879544e80da70099a191613e7e51f",
    },
    "meta": {
        "fname": "TAU-urban-acoustic-scenes-2019-development.meta.zip",
        "url": "https://zenodo.org/record/2589280/files/TAU-urban-acoustic-scenes-2019-development.meta.zip?download=1",
        "hash": "09782f2097e4735687af73c44919329c",
    },
}

MACS_ARCHIVES_FILES = {
    name: TAU_URBAN_ACOUSTIC_DEV_FILES[name]
    for name in (
        "audio.1",
        "audio.10",
        "audio.11",
        "audio.12",
        "audio.13",
        "audio.2",
        "audio.3",
        "audio.9",
    )
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys

from argparse import ArgumentParser, Namespace
from typing import Iterable

import yaml

from aac_datasets.datasets.audiocaps import AudioCaps
from aac_datasets.datasets.clotho import Clotho, CLOTHO_LAST_VERSION
from aac_datasets.datasets.macs import MACS


def to_bool(s: str) -> bool:
    s = s.lower()
    if s in ("true",):
        return True
    elif s in ("false",):
        return False
    else:
        raise ValueError(f"Invalid argument value {s}. (not a boolean)")


def get_main_download_args() -> Namespace:
    parser = ArgumentParser(
        description="Download a dataset at specified root directory."
    )

    parser.add_argument(
        "--root",
        type=str,
        default=".",
        help="The path to the parent directory of the datasets.",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        default=1,
        help="Verbose level of the script. 0 means silent mode, 1 is default mode and 2 add additional debugging outputs.",
    )
    parser.add_argument(
        "--force",
        type=to_bool,
        default=False,
        choices=(False, True),
        help="Force download of files, even if already downloaded.",
    )

    subparsers = parser.add_subparsers(dest="dataset", required=True)

    audiocaps_subparser = subparsers.add_parser("audiocaps")
    audiocaps_subparser.add_argument(
        "--ffmpeg",
        type=str,
        default="ffmpeg",
        help="Path to ffmpeg used to download audio from youtube.",
    )
    audiocaps_subparser.add_argument(
        "--youtube_dl",
        type=str,
        default="youtube-dl",
        help="Path to youtube-dl used to extract metadata from a youtube video.",
    )
    audiocaps_subparser.add_argument(
        "--load_tags",
        type=to_bool,
        default=True,
        choices=(False, True),
        help="Download additional audioset tags corresponding to audiocaps audio.",
    )
    audiocaps_subparser.add_argument(
        "--subsets",
        type=str,
        default=AudioCaps.SUBSETS,
        nargs="+",
        choices=AudioCaps.SUBSETS,
        help="AudioCaps subsets to download.",
    )

    clotho_subparser = subparsers.add_parser("clotho")
    clotho_subparser.add_argument(
        "--version",
        type=str,
        default=CLOTHO_LAST_VERSION,
        choices=Clotho.VERSIONS,
        help="The version of the Clotho dataset.",
    )
    clotho_subparser.add_argument(
        "--clean_archives",
        type=to_bool,
        default=False,
        choices=(False, True),
        help="Remove archives files after extraction.",
    )
    clotho_subparser.add_argument(
        "--subsets",
        type=str,
        default=Clotho.SUBSETS,
        nargs="+",
        choices=Clotho.SUBSETS,
        help="Clotho subsets to download.",
    )

    macs_subparser = subparsers.add_parser("macs")
    macs_subparser.add_argument(
        "--clean_archives",
        type=to_bool,
        default=False,
        choices=(False, True),
        help="Remove archives files after extraction.",
    )

    args = parser.parse_args()
    return args


def download_audiocaps(
    root: str = ".",
    verbose: int = 1,
    force: bool = False,
    ffmpeg: str = "ffmpeg",
    youtube_dl: str = "youtube-dl",
    load_tags: bool = False,
    subsets: Iterable[str] = AudioCaps.SUBSETS,
) -> None:
    """Download AudioCaps dataset subsets."""
    AudioCaps.FORCE_PREPARE_DATA = force
    AudioCaps.FFMPEG_PATH = ffmpeg
    AudioCaps.YOUTUBE_DL_PATH = youtube_dl

    for subset in subsets:
        _ = AudioCaps(root, subset, download=True, verbose=verbose, load_tags=load_tags)


def download_clotho(
    root: str = ".",
    verbose: int = 1,
    force: bool = False,
    version: str = "v2.1",
    clean_archives: bool = False,
    subsets: Iterable[str] = Clotho.SUBSETS,
) -> None:
    """Download Clotho dataset subsets."""
    Clotho.FORCE_PREPARE_DATA = force
    Clotho.CLEAN_ARCHIVES = clean_archives

    for subset in subsets:
        _ = Clotho(root, subset, download=True, verbose=verbose, version=version)


def download_macs(
    root: str = ".",
    verbose: int = 1,
    force: bool = False,
    clean_archives: bool = False,
) -> None:
    """Download MACS dataset."""
    MACS.FORCE_PREPARE_DATA = force
    MACS.CLEAN_ARCHIVES = clean_archives

    _ = MACS(root, download=True, verbose=verbose)


def main_download() -> None:
    format_ = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(format_))
    logger = logging.getLogger("aac_datasets")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    args = get_main_download_args()

    if args.verbose >= 2:
        print(yaml.dump({"Arguments": args.__dict__}, sort_keys=False))

    if args.dataset == "audiocaps":
        download_audiocaps(
            root=args.root,
            verbose=args.verbose,
            force=args.force,
            ffmpeg=args.ffmpeg,
            youtube_dl=args.youtube_dl,
            load_tags=args.load_tags,
            subsets=args.subsets,
        )

    elif args.dataset == "clotho":
        download_clotho(
            root=args.root,
            verbose=args.verbose,
            force=args.force,
            version=args.version,
            clean_archives=args.clean_archives,
            subsets=args.subsets,
        )

    elif args.dataset == "macs":
        download_macs(
            root=args.root,
            verbose=args.verbose,
            force=args.force,
            clean_archives=args.clean_archives,
        )

    else:
        DATASETS = ("audiocaps", "clotho" or "macs")
        raise ValueError(
            f"Invalid argument {args.dataset}. (expected one of {DATASETS})"
        )


if __name__ == "__main__":
    main_download()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import csv
import logging
import os
import os.path as osp

from dataclasses import asdict, astuple, dataclass, field, fields
from functools import cached_property
from typing import Any, Dict, List, Optional, Tuple, Union
from zipfile import ZipFile

import torch
import torchaudio

from py7zr import SevenZipFile
from torch import nn, Tensor
from torch.utils.data.dataset import Dataset
from torchaudio.backend.common import AudioMetaData
from torchaudio.datasets.utils import download_url


logger = logging.getLogger(__name__)


@dataclass
class ClothoItem:
    audio: Tensor = torch.empty((0,))
    captions: List[str] = field(default_factory=list)
    keywords: List[str] = field(default_factory=list)
    fname: Optional[str] = None
    index: Optional[int] = None
    dataset: str = "clotho"
    subset: Optional[str] = None
    sr: int = -1


CLOTHO_LINKS = {
    "v1": {
        "dev": {
            "audio_archive": {
                "fname": "clotho_audio_development.7z",
                "url": "https://zenodo.org/record/3490684/files/clotho_audio_development.7z?download=1",
                "hash_value": "e3ce88561b317cc3825e8c861cae1ec6",
            },
            "captions": {
                "fname": "clotho_captions_development.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_captions_development.csv?download=1",
                "hash_value": "dd568352389f413d832add5cf604529f",
            },
            "metadata": {
                "fname": "clotho_metadata_development.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_metadata_development.csv?download=1",
                "hash_value": "582c18ee47cebdbe33dce1feeab53a56",
            },
        },
        "eval": {
            "audio_archive": {
                "fname": "clotho_audio_evaluation.7z",
                "url": "https://zenodo.org/record/3490684/files/clotho_audio_evaluation.7z?download=1",
                "hash_value": "4569624ccadf96223f19cb59fe4f849f",
            },
            "captions": {
                "fname": "clotho_captions_evaluation.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_captions_evaluation.csv?download=1",
                "hash_value": "1b16b9e57cf7bdb7f13a13802aeb57e2",
            },
            "metadata": {
                "fname": "clotho_metadata_evaluation.csv",
                "url": "https://zenodo.org/record/3490684/files/clotho_metadata_evaluation.csv?download=1",
                "hash_value": "13946f054d4e1bf48079813aac61bf77",
            },
        },
        "test": {
            "audio_archive": {
                "fname": "clotho_audio_test.7z",
                "url": "https://zenodo.org/record/3865658/files/clotho_audio_test.7z?download=1",
                "hash_value": "9b3fe72560a621641ff4351ba1154349",
            },
            "metadata": {
                "fname": "clotho_metadata_test.csv",
                "url": "https://zenodo.org/record/3865658/files/clotho_metadata_test.csv?download=1",
                "hash_value": "52f8ad01c229a310a0ff8043df480e21",
            },
        },
    },
    "v2": {
        "dev": {
            "audio_archive": {
                "fname": "clotho_audio_development.7z",
                "url": "https://zenodo.org/record/4743815/files/clotho_audio_development.7z?download=1",
                "hash_value": "eda144a5e05a60e6d2e37a65fc4720a9",
            },
            "captions": {
                "fname": "clotho_captions_development.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_captions_development.csv?download=1",
                "hash_value": "800633304e73d3daed364a2ba6069827",
            },
            "metadata": {
                "fname": "clotho_metadata_development.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_metadata_development.csv?download=1",
                "hash_value": "5fdc51b4c4f3468ff7d251ea563588c9",
            },
        },
        "eval": {
            "audio_archive": {
                "fname": "clotho_audio_evaluation.7z",
                "url": "https://zenodo.org/record/4743815/files/clotho_audio_evaluation.7z?download=1",
                "hash_value": "4569624ccadf96223f19cb59fe4f849f",
            },
            "captions": {
                "fname": "clotho_captions_evaluation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_captions_evaluation.csv?download=1",
                "hash_value": "1b16b9e57cf7bdb7f13a13802aeb57e2",
            },
            "metadata": {
                "fname": "clotho_metadata_evaluation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_metadata_evaluation.csv?download=1",
                "hash_value": "13946f054d4e1bf48079813aac61bf77",
            },
        },
        "val": {
            "audio_archive": {
                "fname": "clotho_audio_validation.7z",
                "url": "https://zenodo.org/record/4743815/files/clotho_audio_validation.7z?download=1",
                "hash_value": "0475bfa5793e80f748d32525018ebada",
            },
            "captions": {
                "fname": "clotho_captions_validation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_captions_validation.csv?download=1",
                "hash_value": "3109c353138a089c7ba724f27d71595d",
            },
            "metadata": {
                "fname": "clotho_metadata_validation.csv",
                "url": "https://zenodo.org/record/4743815/files/clotho_metadata_validation.csv?download=1",
                "hash_value": "f69cfacebcd47c4d8d30d968f9865475",
            },
        },
        "test": {
            "audio_archive": {
                "fname": "clotho_audio_test.7z",
                "url": "https://zenodo.org/record/3865658/files/clotho_audio_test.7z?download=1",
                "hash_value": "9b3fe72560a621641ff4351ba1154349",
            },
            "metadata": {
                "fname": "clotho_metadata_test.csv",
                "url": "https://zenodo.org/record/3865658/files/clotho_metadata_test.csv?download=1",
                "hash_value": "52f8ad01c229a310a0ff8043df480e21",
            },
        },
    },
    "v2.1": {
        "dev": {
            "audio_archive": {
                "fname": "clotho_audio_development.7z",
                "url": "https://zenodo.org/record/4783391/files/clotho_audio_development.7z?download=1",
                "hash_value": "c8b05bc7acdb13895bb3c6a29608667e",
            },
            "captions": {
                "fname": "clotho_captions_development.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_captions_development.csv?download=1",
                "hash_value": "d4090b39ce9f2491908eebf4d5b09bae",
            },
            "metadata": {
                "fname": "clotho_metadata_development.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_metadata_development.csv?download=1",
                "hash_value": "170d20935ecfdf161ce1bb154118cda5",
            },
        },
        "eval": {
            "audio_archive": {
                "fname": "clotho_audio_evaluation.7z",
                "url": "https://zenodo.org/record/4783391/files/clotho_audio_evaluation.7z?download=1",
                "hash_value": "4569624ccadf96223f19cb59fe4f849f",
            },
            "captions": {
                "fname": "clotho_captions_evaluation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_captions_evaluation.csv?download=1",
                "hash_value": "1b16b9e57cf7bdb7f13a13802aeb57e2",
            },
            "metadata": {
                "fname": "clotho_metadata_evaluation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_metadata_evaluation.csv?download=1",
                "hash_value": "13946f054d4e1bf48079813aac61bf77",
            },
        },
        "val": {
            "audio_archive": {
                "fname": "clotho_audio_validation.7z",
                "url": "https://zenodo.org/record/4783391/files/clotho_audio_validation.7z?download=1",
                "hash_value": "7dba730be08bada48bd15dc4e668df59",
            },
            "captions": {
                "fname": "clotho_captions_validation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_captions_validation.csv?download=1",
                "hash_value": "5879e023032b22a2c930aaa0528bead4",
            },
            "metadata": {
                "fname": "clotho_metadata_validation.csv",
                "url": "https://zenodo.org/record/4783391/files/clotho_metadata_validation.csv?download=1",
                "hash_value": "2e010427c56b1ce6008b0f03f41048ce",
            },
        },
        "test": {
            "audio_archive": {
                "fname": "clotho_audio_test.7z",
                "url": "https://zenodo.org/record/3865658/files/clotho_audio_test.7z?download=1",
                "hash_value": "9b3fe72560a621641ff4351ba1154349",
            },
            "metadata": {
                "fname": "clotho_metadata_test.csv",
                "url": "https://zenodo.org/record/3865658/files/clotho_metadata_test.csv?download=1",
                "hash_value": "52f8ad01c229a310a0ff8043df480e21",
            },
        },
        "analysis": {
            "audio_archive": {
                "fname": "clotho_analysis_2022.zip",
                "url": "https://zenodo.org/record/6610709/files/clotho_analysis_2022.zip?download=1",
                "hash_value": "7e8fa4762cc3a7c5546606680b958d08",
            },
        },
    },
}
CLOTHO_LAST_VERSION = "v2.1"


class Clotho(Dataset):
    """
    Unofficial Clotho pytorch dataset for DCASE 2021 Task 6.
    Subsets available are 'train', 'val', 'eval' and 'test' (for v2 and v2.1).

    Audio are waveform sounds of 15 to 30 seconds, sampled at 44100 Hz.
    Target is a list of 5 different sentences strings describing an audio sample.
    Max number of words in captions is 20.

    Clotho V1 Paper : https://arxiv.org/pdf/1910.09387.pdf

    Dataset folder tree for version 'v2.1':

    ```
    {root}
    └── CLOTHO_v2.1
        ├── clotho_audio_files
        │   ├── clotho_analysis
        │   │    └── (8360 files, ~19GB)
        │   ├── development
        │   │    └── (3840 files, ~7.1GB)
        │   ├── evaluation
        │   │    └── (1045 files, ~2.0GB)
        │   ├── validation
        │   │    └── (1046 files, ~2.0GB)
        │   └── test
        │        └── (1044 files, ~2.0GB)
        └── clotho_csv_files
            ├── clotho_captions_development.csv
            ├── clotho_captions_evaluation.csv
            ├── clotho_captions_validation.csv
            ├── clotho_metadata_development.csv
            ├── clotho_metadata_evaluation.csv
            ├── clotho_metadata_test.csv
            └── clotho_metadata_validation.csv
    ```
    """

    AUDIO_MAX_LENGTH = 30.0  # in seconds
    AUDIO_MIN_LENGTH = 15.0  # in seconds
    AUDIO_N_CHANNELS = 1
    CAPTION_MAX_LENGTH = 20
    CAPTION_MIN_LENGTH = 8
    CAPTIONS_PER_AUDIO = {"dev": 5, "val": 5, "eval": 5, "test": 0}
    CLEAN_ARCHIVES: bool = False
    FORCE_PREPARE_DATA: bool = False
    ITEM_TYPES = ("tuple", "dict", "dataclass")
    SAMPLE_RATE = 44100
    SUBSETS_DICT = {
        version: tuple(links.keys()) for version, links in CLOTHO_LINKS.items()
    }
    SUBSETS = SUBSETS_DICT[CLOTHO_LAST_VERSION]
    VERSIONS = tuple(CLOTHO_LINKS.keys())

    def __init__(
        self,
        root: str = ".",
        subset: str = "dev",
        download: bool = False,
        transforms: Optional[Dict[str, Optional[nn.Module]]] = None,
        unfold: bool = False,
        item_type: str = "tuple",
        version: str = "v2.1",
        verbose: int = 0,
    ) -> None:
        """
        :param root: The parent of the dataset root directory.
            Note: The data is stored in the 'CLOTHO_{version}' subdirectory.
            defaults to ".".
        :param subset: The subset of Clotho to use.
            Can be 'dev', 'val', 'eval', 'test' or 'analysis' for clotho v2.1.
        :param download: Download the dataset if download=True and if the dataset is not already downloaded.
            default to False.
        :param transforms: The transform to apply values (Tensor).
            Keys can be 'audio', 'captions' or 'keywords'.
            defaults to None.
        :param unfold: If True, map captions to audio instead of audio to caption.
            defaults to True.
        :param item_type: The type of the value returned by __getitem__.
            Can be 'tuple', 'dict' or 'dataclass'.
            defaults to 'tuple'.
        :param version: The version of the dataset. Can be 'v1', 'v2' or 'v2.1'.
            defaults to 'v2.1'.
        :param verbose: Verbose level to use. Can be 0 or 1.
            defaults to 0.
        """
        if version not in Clotho.VERSIONS:
            raise ValueError(
                f"Invalid Clotho argument {version=}. Must be one of {Clotho.VERSIONS}."
            )

        if subset not in CLOTHO_LINKS[version].keys():
            raise ValueError(
                f"Invalid Clotho argument {subset=} for {version=}. Must be one of {tuple(CLOTHO_LINKS[version].keys())}."
            )

        if item_type not in self.ITEM_TYPES:
            raise ValueError(
                f"Invalid argument {item_type=} for Clotho. (expected one of {self.ITEM_TYPES})"
            )

        if transforms is None:
            transforms = {}

        accepted_keys = [field_.name for field_ in fields(ClothoItem)]
        for key in transforms.keys():
            if key not in accepted_keys:
                raise ValueError(
                    f"Invalid transform {key=}. (expected one of {accepted_keys})"
                )

        super().__init__()
        self._root = root
        self._subset = subset
        self._download = download
        self._transforms = transforms
        self._unfold = unfold
        self._item_type = item_type
        self._version = version
        self._verbose = verbose

        self._all_infos = []
        self._metadata_keys = []

        if self._download:
            self._prepare_data()
        self._load_data()

    def __getitem__(self, index: int) -> Union[Tuple, Dict, ClothoItem]:
        """
        Get the audio data as 1D tensor and the matching captions as 5 sentences.

        :param index: The index of the item.
        :return: Audio data of shape (size,) and the 5 matching captions as tuple or dict.
        """
        kwargs = {
            field.name: self.get(field.name, index) for field in fields(ClothoItem)
        }
        item = ClothoItem(**kwargs)

        if self._item_type == "tuple":
            return astuple(item)
        elif self._item_type == "dict":
            return asdict(item)
        elif self._item_type == "dataclass":
            return item
        else:
            raise ValueError(
                f"Invalid item_type={self._item_type} for Clotho_{self._subset}. (expected one of {self.ITEM_TYPES})"
            )

    def __len__(self) -> int:
        """
        :return: The number of items in the dataset.
        """
        return len(self._all_infos)

    def get_valid_item_keys(self) -> Tuple[str, ...]:
        return ("audio", "fpath", "dataset", "subset", "index") + tuple(
            self._all_infos[0].keys()
        )

    def get_raw(self, name: str, index: int) -> Any:
        """Read the processed raw data. (without transform)

        :param name: The name of the value.
                Can be 'audio', 'fpath', 'fname', 'captions', 'keywords', 'sound_id', 'sound_link', 'start_end_samples', 'manufacturer', 'license'.
        :param index: The index of the value in range [0, len(dataset)[.
        """
        if not (0 <= index < len(self)):
            raise IndexError(
                f"Invalid argument {index=} for {self} (expected in range [0, {len(self)}-1])"
            )

        if name == "audio":
            fpath = self.get("fpath", index)
            value, sr = torchaudio.load(fpath)  # type: ignore

            # Sanity check
            if sr != self.SAMPLE_RATE:
                raise RuntimeError(
                    f"Invalid sample rate {sr}Hz of audio file {fpath} with Clotho {self.SAMPLE_RATE}Hz."
                )

        elif name == "fpath":
            fname = self.get("fname", index)
            value = osp.join(self._dpath_audio_subset, fname)

        elif name == "dataset":
            value = "clotho"

        elif name == "subset":
            value = self._subset

        elif name == "index":
            value = index

        elif name == "sr":
            fpath = self.get("fpath", index)
            audio_info: AudioMetaData = torchaudio.info(fpath)  # type: ignore
            return audio_info.sample_rate

        elif name in self._all_infos[index].keys():
            value = self._all_infos[index][name]

        else:
            raise ValueError(
                f"Invalid value {name=} at {index=}. (dataset=clotho, subset={self._subset}, len={len(self)})"
            )
        return value

    def get(self, name: str, index: int) -> Any:
        """Read the processed data. (with transform)

        :param name: The name of the value.
                Can be 'audio', 'fpath', 'fname', 'captions', 'keywords', 'sound_id', 'sound_link', 'start_end_samples', 'manufacturer', 'license'.
        :param index: The index of the value in range [0, len(dataset)[.
        """
        value = self.get_raw(name, index)
        transform = self._transforms.get(name, None)
        if transform is not None:
            value = transform(value)
        return value

    @cached_property
    def _dpath_data(self) -> str:
        return osp.join(self._root, f"CLOTHO_{self._version}")

    @cached_property
    def _dpath_audio(self) -> str:
        return osp.join(self._dpath_data, "clotho_audio_files")

    @cached_property
    def _dpath_audio_subset(self) -> str:
        return osp.join(
            self._dpath_data,
            "clotho_audio_files",
            CLOTHO_AUDIO_DNAMES[self._subset],
        )

    @cached_property
    def _dpath_csv(self) -> str:
        return osp.join(self._dpath_data, "clotho_csv_files")

    def _is_prepared(self) -> bool:
        if not all(map(osp.isdir, (self._dpath_audio_subset, self._dpath_csv))):
            return False

        if self._subset in ("test", "analysis"):
            return True

        links = CLOTHO_LINKS[self._version][self._subset]
        captions_fname = links["captions"]["fname"]
        captions_fpath = osp.join(self._dpath_csv, captions_fname)
        with open(captions_fpath, "r") as file:
            reader = csv.DictReader(file)
            lines = list(reader)
        return len(lines) == len(os.listdir(self._dpath_audio_subset))

    def _prepare_data(self) -> None:
        if not osp.isdir(self._root):
            raise RuntimeError(f"Cannot find root directory '{self._root}'.")

        os.makedirs(self._dpath_audio, exist_ok=True)
        os.makedirs(self._dpath_csv, exist_ok=True)

        if self._verbose >= 1:
            logger.info(f"Start to download files for clotho_{self._subset}...")

        links = copy.deepcopy(CLOTHO_LINKS[self._version][self._subset])
        extensions = ("7z", "csv", "zip")

        # Download csv and 7z files
        for info in links.values():
            fname, url, hash_value = info["fname"], info["url"], info["hash_value"]
            extension = fname.split(".")[-1]

            if extension in ("7z", "zip"):
                dpath = self._dpath_audio
            elif extension == "csv":
                dpath = self._dpath_csv
            else:
                raise RuntimeError(
                    f"Found invalid {extension=}. Must be one of {extensions}."
                )

            fpath = osp.join(dpath, fname)
            if not osp.isfile(fpath) or self.FORCE_PREPARE_DATA:
                if self._verbose >= 1:
                    logger.info(f"Download and check file '{fname}' from {url=}...")

                download_url(
                    url,
                    dpath,
                    fname,
                    hash_value=hash_value,
                    hash_type="md5",
                    progress_bar=self._verbose >= 1,
                )

            elif self._verbose >= 1:
                logger.info(f"File {fname=} is already extracted.")

        # Extract audio files from archives
        for info in links.values():
            fname = info["fname"]
            extension = fname.split(".")[-1]

            if extension in ("7z", "zip"):
                fpath = osp.join(self._dpath_audio, fname)

                if self._verbose >= 1:
                    logger.info(f"Extract archive file {fname=}...")

                if extension == "7z":
                    archive_file = SevenZipFile(fpath)
                    compressed_fnames = [
                        osp.basename(fname) for fname in archive_file.getnames()
                    ]
                elif extension == "zip":
                    archive_file = ZipFile(fpath)
                    compressed_fnames = [
                        osp.basename(file.filename) for file in archive_file.filelist
                    ]
                else:
                    raise RuntimeError

                # Ignore dir name from archive file
                compressed_fnames = [
                    fname
                    for fname in compressed_fnames
                    if fname not in ("", CLOTHO_AUDIO_DNAMES[self._subset])
                ]
                extracted_fnames = (
                    os.listdir(self._dpath_audio_subset)
                    if osp.isdir(self._dpath_audio_subset)
                    else []
                )

                if set(extracted_fnames) != set(compressed_fnames):
                    archive_file.extractall(self._dpath_audio)

                    # Check if files is good now
                    extracted_fnames = os.listdir(self._dpath_audio_subset)
                    if set(extracted_fnames) != set(compressed_fnames):
                        raise RuntimeError(
                            f"Invalid number of audios extracted. (found {len(extracted_fnames)} files but expected the same {len(compressed_fnames)} files)"
                        )

                archive_file.close()

            elif extension == "csv":
                pass

            else:
                logger.error(
                    f"Found unexpected {extension=} for downloaded file '{fname}'. Expected one of {extensions}."
                )

        if self.CLEAN_ARCHIVES:
            for info in links.values():
                fname = info["fname"]
                extension = fname.split(".")[-1]

                if extension in ("7z", "zip"):
                    fpath = osp.join(self._dpath_audio, fname)
                    if self._verbose >= 1:
                        logger.info(f"Removing archive file {osp.basename(fpath)}...")
                    os.remove(fpath)

    def _load_data(self) -> None:
        if not self._is_prepared():
            raise RuntimeError(
                f"Cannot load data: clotho_{self._subset} is not prepared in data root={self._root}. Please use download=True in dataset constructor."
            )

        # Read fpath of .wav audio files
        links = CLOTHO_LINKS[self._version][self._subset]
        dpath_audio_subset = self._dpath_audio_subset

        if not osp.isdir(dpath_audio_subset):
            raise RuntimeError(f'Cannot find directory "{dpath_audio_subset}".')

        self._all_infos = [
            {
                "fname": fname,
                "captions": [],
                "keywords": [],
            }
            for fname in os.listdir(dpath_audio_subset)
        ]
        idx_to_fname = {i: infos["fname"] for i, infos in enumerate(self._all_infos)}
        fname_to_idx = {fname: i for i, fname in idx_to_fname.items()}

        # --- Read captions info
        if "captions" in links.keys():
            captions_fname = links["captions"]["fname"]
            captions_fpath = osp.join(self._dpath_csv, captions_fname)

            with open(captions_fpath, "r") as file:
                reader = csv.DictReader(file)
                for row in reader:
                    # Keys: file_name, caption_1, caption_2, caption_3, caption_4, caption_5
                    fname = row["file_name"]

                    if fname not in fname_to_idx.keys():
                        raise RuntimeError(
                            f"Found '{fname=}' in CSV '{captions_fname=}' but not the corresponding audio file."
                        )
                    else:
                        idx = fname_to_idx[fname]
                        self._all_infos[idx]["captions"] = list(
                            row[caption_key]
                            for caption_key in (
                                "caption_1",
                                "caption_2",
                                "caption_3",
                                "caption_4",
                                "caption_5",
                            )
                        )

        # --- Read metadata info
        if "metadata" in links.keys():
            metadata_fname = links["metadata"]["fname"]
            metadata_fpath = osp.join(self._dpath_csv, metadata_fname)

            # Keys: file_name, keywords, sound_id, sound_link, start_end_samples, manufacturer, license
            if self._version in ("v2", "v2.1"):
                encoding = "ISO-8859-1"
            else:
                encoding = None

            with open(metadata_fpath, "r", encoding=encoding) as file:
                delimiter = ";" if self._subset == "test" else ","
                reader = csv.DictReader(file, delimiter=delimiter)
                self._metadata_keys = (
                    reader.fieldnames if reader.fieldnames is not None else ()
                )

                for row in reader:
                    # file_name,keywords,sound_id,sound_link,start_end_samples,manufacturer,license
                    fname = row["file_name"]
                    row_copy: Any = copy.deepcopy(row)
                    row_copy.pop("file_name")

                    if "keywords" in row_copy.keys():
                        row_copy["keywords"] = row_copy["keywords"].split(";")
                    else:
                        row_copy["keywords"] = []

                    if fname in fname_to_idx.keys():
                        idx = fname_to_idx[fname]
                        self._all_infos[idx].update(row_copy)
                    else:
                        raise RuntimeError(
                            f"Found value {fname=} in CSV metadata file '{metadata_fname}' but not the corresponding audio file."
                        )

        if self._unfold and self.CAPTIONS_PER_AUDIO[self._subset] > 1:
            data_info_unfolded = []
            for infos in self._all_infos:
                captions = infos["captions"]
                for caption in captions:
                    new_infos = dict(infos)
                    new_infos["captions"] = [caption]
                    data_info_unfolded.append(new_infos)

            self._all_infos = data_info_unfolded

        self._all_infos = self._all_infos
        if self._verbose >= 1:
            logger.info(f"clotho_{self._subset} has been loaded. (len={len(self)})")

    def __repr__(self) -> str:
        return f"Clotho(subset={self._subset})"


CLOTHO_AUDIO_DNAMES = {
    "dev": "development",
    "eval": "evaluation",
    "test": "test",
    "val": "validation",
    "analysis": "clotho_analysis",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import sys

from typing import Dict

import torch
import torchaudio
import yaml

import aac_datasets


def get_packages_versions() -> Dict[str, str]:
    return {
        "aac_datasets": aac_datasets.__version__,
        "python": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        "os": platform.system(),
        "architecture": platform.architecture()[0],
        "torch": str(torch.__version__),
        "torchaudio": torchaudio.__version__,
    }


def main_version(*args, **kwargs) -> None:
    """Print some packages versions."""
    versions = get_packages_versions()
    print(yaml.dump(versions, sort_keys=False))


if __name__ == "__main__":
    main_version()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import logging
import os
import os.path as osp
import subprocess
import sys
import time
import tqdm

from dataclasses import asdict, astuple, dataclass, field, fields
from subprocess import CalledProcessError
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torchaudio

from torch import nn, Tensor
from torch.utils.data.dataset import Dataset
from torchaudio.backend.common import AudioMetaData
from torchaudio.datasets.utils import download_url


logger = logging.getLogger(__name__)


@dataclass
class AudioCapsItem:
    audio: Tensor = torch.empty((0,), dtype=torch.float)
    captions: List[str] = field(default_factory=list)
    tags: Optional[List[str]] = None
    fname: Optional[str] = None
    index: int = -1
    dataset: str = "audiocaps"
    subset: Optional[str] = None
    sr: int = -1


class AudioCaps(Dataset):
    """
    Unofficial AudioCaps pytorch dataset.
    Subsets available are 'train', 'val' and 'test'.

    Audio is a waveform tensor of shape (1, n_times) of 10 seconds max, sampled at 16 KHz.
    Target is a list of strings containing the captions.
    The 'train' subset has only 1 caption per sample and 'val' and 'test' have 5 captions.

    Download requires 'youtube-dl' and 'ffmpeg' commands.
    You can change the default path with 'AudioCaps.YOUTUBE_DL_PATH' or 'AudioCaps.FFMPEG_PATH' global variables.

    AudioCaps paper : https://www.aclweb.org/anthology/N19-1011.pdf

    Dataset folder tree :

    ```
    root/
    └── AUDIOCAPS_32000Hz
        ├── train.csv
        ├── val.csv
        ├── test.csv
        └── audio
            ├── train
            │    └── (46231/49838 files, ~42G for 32KHz)
            ├── val
            │    └── (465/495 files, ~425M for 32KHz)
            └── test
                └── (913/975 files, ~832M for 32KHz)
    ```
    """

    AUDIO_FILE_EXTENSION = "flac"
    AUDIO_N_CHANNELS = 1
    AUDIO_MAX_LENGTH = 10.00096876  # in seconds
    AUDIO_MIN_LENGTH = 0.6501874  # in seconds
    CAPTION_MIN_LENGTH = 1
    CAPTION_MAX_LENGTH = 52
    CAPTIONS_PER_AUDIO = {"train": 1, "val": 5, "test": 5}
    DNAME_AUDIO = "audio"
    DNAME_LOG = "logs"
    FFMPEG_PATH: str = "ffmpeg"
    FORCE_PREPARE_DATA: bool = False
    ITEM_TYPES = ("tuple", "dict", "dataclass")
    REDIRECT_LOG = False
    SAMPLE_RATE = 32000
    SUBSETS = ("train", "val", "test")
    VERIFY_FILES = False
    YOUTUBE_DL_PATH: str = "youtube-dl"

    def __init__(
        self,
        root: str = ".",
        subset: str = "train",
        download: bool = False,
        transforms: Optional[Dict[str, Optional[nn.Module]]] = None,
        accept_invalid_fpath: bool = False,
        unfold: bool = False,
        item_type: str = "tuple",
        load_tags: bool = True,
        verbose: int = 0,
    ) -> None:
        """
        :param root: Dataset root directory.
            The data will be stored in the 'AUDIOCAPS_{SAMPLE_RATE}' subdirectory.
            defaults to ".".
        :param subset: The AudioCaps subset.
            Can be 'train', 'val' or 'test'.
        :param download: If True, starts to download the dataset from Youtube. Can requires a lot of time depending of your machine.
            defaults to False.
        :param transforms: The dictionary of transforms to apply.
            All keys must be a valid item attribute, i.e. one of "audio", "captions", "tags", "fname", "index", "dataset", "subset", "fpath", "youtube_id", "start_time", "sr".
            default to None.
        :param accept_invalid_fpath: If True, the dataset will return an empty Tensor when the audio is not present on the disk.
            defaults to False.
        :param unfold: If True, map captions to audio instead of audio to caption.
            defaults to True.
        :param item_type: The type of the value returned by __getitem__.
            Can be 'tuple', 'dict' or 'dataclass'.
            defaults to 'tuple'.
        :param load_tags: If True, load the tags from AudioSet dataset.
            Note: tags needs to be downloaded with download=True & load_tags=True before being used.
            defaults to False.
        :param verbose: Verbose level.
            defaults to 0.
        """
        if subset not in self.SUBSETS:
            raise ValueError(
                f"Invalid argument {subset=} for AudioCaps. (expected one of {self.SUBSETS})"
            )

        if item_type not in self.ITEM_TYPES:
            raise ValueError(
                f"Invalid argument {item_type=} for AudioCaps. (expected one of {self.ITEM_TYPES})"
            )

        if transforms is None:
            transforms = {}

        accepted_keys = [field_.name for field_ in fields(AudioCapsItem)]
        for key in transforms.keys():
            if key not in accepted_keys:
                raise ValueError(
                    f"Invalid argument transforms {key=}. (expected one of {accepted_keys})"
                )

        super().__init__()
        self._root = root
        self._subset = subset
        self._download = download
        self._transforms = transforms
        self._accept_invalid_fpath = accept_invalid_fpath
        self._unfold = unfold
        self._item_type = item_type
        self._load_tags = load_tags
        self._verbose = verbose

        self._all_infos = []
        self._data_dname = f"AUDIOCAPS_{self.SAMPLE_RATE}Hz"

        if self._download:
            self._prepare_data()
        self._load_data()

    def __len__(self) -> int:
        return len(self._all_infos)

    def __getitem__(self, index: int) -> Union[Tuple, Dict, AudioCapsItem]:
        """
        Get the audio data as 1D tensor and the matching captions.

        :param index: The index of the value in range [0, len(dataset)[.
        :return: Item containing the following data: "audio", "captions", "tags", "fname", "index", "dataset", "subset".
            The item type is determined by the 'item_type' attribute of the dataset, and can be a tuple, dict or AudioCapsItem.
        """
        kwargs = {
            field.name: self.get(field.name, index) for field in fields(AudioCapsItem)
        }
        item = AudioCapsItem(**kwargs)

        if self._item_type == "tuple":
            return astuple(item)
        elif self._item_type == "dict":
            return asdict(item)
        elif self._item_type == "dataclass":
            return item
        else:
            raise ValueError(
                f"Invalid item_type={self._item_type} for AudioCaps_{self._subset}. (expected one of {self.ITEM_TYPES})"
            )

    def get_raw(self, name: str, index: int) -> Any:
        """Read a raw data. (without transform applied)

        :param name: The name of the value.
            Can be "audio", "captions", "tags", "fname", "index", "dataset", "subset", "fpath", "youtube_id", "start_time", "sr".
        :param index: The index of the value in range [0, len(dataset)[.
        :return: The value of the data 'name' at specified index.
        """
        if not (0 <= index < len(self)):
            raise IndexError(
                f"Invalid argument {index=} for {self} (expected in range [0, {len(self)}-1])"
            )

        if name == "audio":
            fpath = self.get("fpath", index)

            if fpath is None and self._accept_invalid_fpath:
                return torch.empty((0,), dtype=torch.float)

            value, sr = torchaudio.load(fpath)  # type: ignore
            if sr != self.SAMPLE_RATE:
                raise RuntimeError(
                    f"Invalid sample rate {sr}Hz of audio file {fpath} with AudioCaps {self.SAMPLE_RATE}Hz."
                )

        elif name == "fpath":
            audio_subset_dpath = osp.join(
                self._root, self._data_dname, self.DNAME_AUDIO, self._subset
            )
            fname = self.get("fname", index)
            if fname is None:
                value = None
            else:
                value = osp.join(audio_subset_dpath, fname)

        elif name == "dataset":
            value = "audiocaps"

        elif name == "subset":
            value = self._subset

        elif name == "index":
            value = index

        elif name == "sr":
            fpath = self.get("fpath", index)

            if fpath is None and self._accept_invalid_fpath:
                return -1

            audio_info: AudioMetaData = torchaudio.info(fpath)  # type: ignore
            return audio_info.sample_rate

        elif name in self._all_infos[index].keys():
            # keys : "youtube_id", "fname", "start_time", "captions", "tags"
            value = self._all_infos[index][name]

        else:
            raise ValueError(f"Invalid value {name=} at {index=}.")
        return value

    def get(self, name: str, index: int) -> Any:
        """Read a processed data. (with transform)

        :param name: The name of the value.
            Can be "audio", "captions", "tags", "fname", "index", "dataset", "subset", "fpath", "youtube_id", "start_time", "sr".
        :param index: The index of the value in range [0, len(dataset)[.
        :return: The value of the data 'name' at specified index.
        """
        value = self.get_raw(name, index)
        transform = self._transforms.get(name, None)
        if transform is not None:
            value = transform(value)
        return value

    def _is_prepared(self) -> bool:
        root_dpath = osp.join(self._root, self._data_dname)
        audio_subset_dpath = osp.join(root_dpath, self.DNAME_AUDIO, self._subset)
        links = AUDIOCAPS_LINKS[self._subset]
        meta_fname = links["meta"]["fname"]
        meta_fpath = osp.join(root_dpath, meta_fname)
        return osp.isdir(audio_subset_dpath) and osp.isfile(meta_fpath)

    def _prepare_data(self) -> None:
        if not osp.isdir(self._root):
            raise RuntimeError(f"Cannot find root directory '{self._root}'.")

        try:
            subprocess.check_call(
                [self.YOUTUBE_DL_PATH, "--help"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except (CalledProcessError, PermissionError, FileNotFoundError) as err:
            logger.error(
                f"Cannot use youtube-dl path '{self.YOUTUBE_DL_PATH}'. ({err})"
            )
            raise err

        try:
            subprocess.check_call(
                [self.FFMPEG_PATH, "--help"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except (CalledProcessError, PermissionError, FileNotFoundError) as err:
            logger.error(f"Cannot use ffmpeg path '{self.FFMPEG_PATH}'. ({err})")
            raise err

        if self._is_prepared() and not self.FORCE_PREPARE_DATA:
            return None

        links = AUDIOCAPS_LINKS[self._subset]
        root_dpath = osp.join(self._root, self._data_dname)
        audio_subset_dpath = osp.join(root_dpath, self.DNAME_AUDIO, self._subset)
        meta_fname = links["meta"]["fname"]
        meta_fpath = osp.join(root_dpath, meta_fname)

        if not osp.isdir(audio_subset_dpath):
            os.makedirs(audio_subset_dpath)

        if not osp.isfile(meta_fpath):
            url_meta = links["meta"]["url"]
            download_url(url_meta, root_dpath, progress_bar=self._verbose >= 1)

        start = time.perf_counter()
        with open(meta_fpath, "r") as file:
            n_samples = len(file.readlines())

        if self._verbose >= 1:
            log_dpath = osp.join(root_dpath, self.DNAME_LOG)
            if not osp.isdir(log_dpath):
                os.makedirs(log_dpath)

            if self.REDIRECT_LOG:
                logging.basicConfig(
                    filename=osp.join(log_dpath, f"preparation_{self._subset}.txt"),
                    filemode="w",
                    level=logging.INFO,
                    force=True,
                )
            logger.info(f"Start downloading files for {self._subset} AudioCaps split.")

        with open(meta_fpath, "r") as file:
            # Download audio files
            reader = csv.DictReader(file)
            if self._verbose >= 1:
                reader = tqdm.tqdm(reader, total=n_samples)

            n_download_ok, n_download_err, n_already_ok, n_already_err = 0, 0, 0, 0
            for line in reader:
                # Keys: audiocap_id, youtube_id, start_time, caption
                audiocap_id, youtube_id, start_time = [
                    line[key] for key in ("audiocap_id", "youtube_id", "start_time")
                ]
                fpath = osp.join(
                    audio_subset_dpath,
                    f"{youtube_id}_{start_time}.{self.AUDIO_FILE_EXTENSION}",
                )
                if not start_time.isdigit():
                    raise RuntimeError(
                        f'Start time "{start_time}" is not an integer (audiocap_id={audiocap_id}, youtube_id={youtube_id}).'
                    )
                start_time = int(start_time)

                if not osp.isfile(fpath):
                    success = download_and_extract_from_youtube(
                        youtube_id,
                        fpath,
                        start_time,
                        duration=self.AUDIO_MAX_LENGTH,
                        sr=self.SAMPLE_RATE,
                        youtube_dl_path=self.YOUTUBE_DL_PATH,
                        ffmpeg_path=self.FFMPEG_PATH,
                        n_channels=self.AUDIO_N_CHANNELS,
                    )
                    if success:
                        valid_file = self._check_file(fpath)
                        if valid_file:
                            if self._verbose >= 2:
                                logger.debug(
                                    f'[{audiocap_id:6s}] File "{youtube_id}" has been downloaded and verified.'
                                )
                            n_download_ok += 1
                        else:
                            if self._verbose >= 1:
                                logger.warning(
                                    f'[{audiocap_id:6s}] File "{youtube_id}" has been downloaded but it is not valid and it will be removed.'
                                )
                            os.remove(fpath)
                            n_download_err += 1
                    else:
                        logger.error(
                            f'[{audiocap_id:6s}] Cannot extract audio from "{youtube_id}".'
                        )
                        n_download_err += 1

                elif self.VERIFY_FILES:
                    valid_file = self._check_file(fpath)
                    if valid_file:
                        if self._verbose >= 2:
                            logger.debug(
                                f'[{audiocap_id:6s}] File "{youtube_id}" is already downloaded and has been verified.'
                            )
                        n_already_ok += 1
                    else:
                        if self._verbose >= 1:
                            logger.warning(
                                f'[{audiocap_id:6s}] File "{youtube_id}" is already downloaded but it is not valid and will be removed.'
                            )
                        os.remove(fpath)
                        n_already_err += 1
                else:
                    if self._verbose >= 2:
                        logger.debug(
                            f'[{audiocap_id:6s}] File "{youtube_id}" is already downloaded but it is not verified due to {self.VERIFY_FILES=}.'
                        )
                    n_already_ok += 1

        if self._load_tags:
            for key in ("class_labels_indices", "unbalanced"):
                infos = AUDIOSET_LINKS[key]
                url = infos["url"]
                fname = infos["fname"]
                fpath = osp.join(root_dpath, fname)
                if not osp.isfile(fpath):
                    if self._verbose >= 1:
                        logger.info(f"Downloading file '{fname}'...")
                    download_url(url, root_dpath, fname, progress_bar=self._verbose > 0)

        if self._verbose >= 1:
            duration = int(time.perf_counter() - start)
            logger.info(
                f'Download and preparation of AudioCaps for subset "{self._subset}" done in {duration}s. '
            )
            logger.info(f"- {n_download_ok} downloads success,")
            logger.info(f"- {n_download_err} downloads failed,")
            logger.info(f"- {n_already_ok} already downloaded,")
            logger.info(f"- {n_already_err} already downloaded errors,")
            logger.info(f"- {n_samples} total samples.")

            if self.REDIRECT_LOG:
                logging.basicConfig(
                    stream=sys.stdout,
                    level=logging.INFO,
                    force=True,
                )

    def _check_file(self, fpath: str) -> bool:
        try:
            audio, sr = torchaudio.load(fpath)  # type: ignore
        except RuntimeError:
            message = f'Found file "{fpath}" already downloaded but it is invalid (cannot load). It will be removed.'
            logger.error(message)
            return False

        if audio.nelement() == 0:
            message = f'Found file "{fpath}" already downloaded but it is invalid (empty audio). It will be removed.'
            logger.error(message)
            return False

        if sr != self.SAMPLE_RATE:
            message = f'Found file "{fpath}" already downloaded but it is invalid (invalid sr={sr} != {self.SAMPLE_RATE}). It will be removed.'
            logger.error(message)
            return False

        return True

    def _load_data(self) -> None:
        if not self._is_prepared():
            raise RuntimeError(
                f"Cannot load data: audiocaps_{self._subset} is not prepared in data root={self._root}. Please use download=True in dataset constructor."
            )

        links = AUDIOCAPS_LINKS[self._subset]
        root_dpath = osp.join(self._root, self._data_dname)
        audio_subset_dpath = osp.join(root_dpath, self.DNAME_AUDIO, self._subset)
        meta_fname = links["meta"]["fname"]
        meta_fpath = osp.join(root_dpath, meta_fname)

        youtube_id_to_index = {}
        self._all_infos = []

        with open(meta_fpath, "r") as file:
            reader = csv.DictReader(file)

            for line_dict in tqdm.tqdm(
                reader,
                disable=self._verbose == 0,
                desc=f"audiocaps_{self._subset}: Loading captions...",
            ):
                youtube_id = line_dict["youtube_id"]
                start_time = line_dict["start_time"]
                caption = line_dict["caption"]

                fname = f"{youtube_id}_{start_time}.{self.AUDIO_FILE_EXTENSION}"
                fpath = osp.join(audio_subset_dpath, fname)
                isfile = osp.isfile(fpath)

                if isfile or self._accept_invalid_fpath:
                    if not isfile:
                        fname = None

                    if youtube_id in youtube_id_to_index.keys():
                        index = youtube_id_to_index[youtube_id]
                        captions = self._all_infos[index]["captions"]
                        assert fname == self._all_infos[index]["fname"]
                        assert start_time == self._all_infos[index]["start_time"]
                        self._all_infos[index]["captions"].append(caption)

                    else:
                        new_index = len(self._all_infos)
                        youtube_id_to_index[youtube_id] = new_index
                        info = {
                            "youtube_id": youtube_id,
                            "fname": fname,
                            "start_time": start_time,
                            "captions": [caption],
                            "tags": None,
                        }
                        self._all_infos.append(info)

        if self._load_tags:
            class_labels_indices_fpath = osp.join(
                root_dpath, AUDIOSET_LINKS["class_labels_indices"]["fname"]
            )
            unbalanced_tags_fpath = osp.join(
                root_dpath, AUDIOSET_LINKS["unbalanced"]["fname"]
            )

            mid_to_tag_name = {}
            tag_name_to_index = {}

            with open(class_labels_indices_fpath, "r") as file:
                reader = csv.DictReader(file)
                # index,mid,display_name
                for line in reader:
                    mid_to_tag_name[line["mid"]] = line["display_name"]
                    tag_name_to_index[line["display_name"]] = int(line["index"])

            with open(unbalanced_tags_fpath, "r") as file:
                fieldnames = ("YTID", "start_seconds", "end_seconds", "positive_labels")
                reader = csv.DictReader(
                    file, fieldnames, skipinitialspace=True, strict=True
                )
                # Skip the comments
                for _ in range(3):
                    next(reader)

                # YTID, start_seconds, end_seconds, positive_labels
                for line in tqdm.tqdm(
                    reader, disable=self._verbose <= 0, desc="Loading tags..."
                ):
                    youtube_id = line["YTID"]
                    if youtube_id in youtube_id_to_index.keys():
                        index = youtube_id_to_index[youtube_id]

                        # Note : In audiocaps, start_time is a string repr of an integer value
                        start_time_audiocaps = float(
                            self._all_infos[index]["start_time"]
                        )
                        # Note : In audioset, start_time is a string repr of a float value
                        start_time_audioset = float(line["start_seconds"])
                        assert start_time_audiocaps == start_time_audioset

                        tags = line["positive_labels"]
                        tags = tags.split(",")
                        tags = [mid_to_tag_name[tag] for tag in tags]
                        self._all_infos[index]["tags"] = tags

            assert all(
                "tags" in info.keys() and info["tags"] is not None
                for info in self._all_infos
            )

        if self._unfold and self.CAPTIONS_PER_AUDIO[self._subset] > 1:
            data_info_unfolded = []
            for links in self._all_infos:
                captions = links["captions"]
                for caption in captions:
                    new_infos = dict(links)
                    new_infos["captions"] = (caption,)
                    data_info_unfolded.append(new_infos)

            self._all_infos = data_info_unfolded

        self._all_infos = self._all_infos
        if self._verbose >= 1:
            logger.info(
                f"{self.__class__.__name__}/{self._subset} has been loaded. (len={len(self)})"
            )

    def __repr__(self) -> str:
        return f"AudioCaps(subset={self._subset})"


def download_and_extract_from_youtube(
    youtube_id: str,
    fpath_out: str,
    start_time: int,
    duration: float = 10.0,
    sr: int = 16000,
    n_channels: int = 1,
    target_format: str = "flac",
    acodec: str = "flac",
    youtube_dl_path: str = "youtube-dl",
    ffmpeg_path: str = "ffmpeg",
) -> bool:
    # Get audio download link with youtube-dl
    link = f"https://www.youtube.com/watch?v={youtube_id}"
    get_url_command = [
        youtube_dl_path,
        "--youtube-skip-dash-manifest",
        "-g",
        link,
    ]
    try:
        output = subprocess.check_output(get_url_command)
    except (CalledProcessError, PermissionError):
        return False

    output = output.decode()
    lines = output.split("\n")
    if len(lines) < 2:
        return False
    _video_link, audio_link = lines[:2]

    # Download and extract audio from audio_link to fpath_out with ffmpeg
    extract_command = [
        ffmpeg_path,
        # Input
        "-i",
        audio_link,
        # Remove video
        "-vn",
        # Format (flac)
        "-f",
        target_format,
        # Audio codec (flac)
        "-acodec",
        acodec,
        # Get only 10s of the clip after start_time
        "-ss",
        str(start_time),
        "-t",
        str(duration),
        # Resample to 16 KHz
        "-ar",
        str(sr),
        # Compute mean of 2 channels
        "-ac",
        str(n_channels),
        fpath_out,
    ]
    try:
        exitcode = subprocess.check_call(extract_command)
        return exitcode == 0
    except (CalledProcessError, PermissionError):
        return False


AUDIOCAPS_LINKS = {
    "train": {
        "meta": {
            "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset/train.csv",
            "fname": "train.csv",
        },
    },
    "val": {
        "meta": {
            "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset/val.csv",
            "fname": "val.csv",
        },
    },
    "test": {
        "meta": {
            "url": "https://raw.githubusercontent.com/cdjkim/audiocaps/master/dataset/test.csv",
            "fname": "test.csv",
        },
    },
}

AUDIOSET_LINKS = {
    "class_labels_indices": {
        "fname": "class_labels_indices.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/class_labels_indices.csv",
    },
    "eval": {
        "fname": "eval_segments.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/eval_segments.csv",
    },
    "balanced": {
        "fname": "balanced_train_segments.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/balanced_train_segments.csv",
    },
    "unbalanced": {
        "fname": "unbalanced_train_segments.csv",
        "url": "http://storage.googleapis.com/us_audioset/youtube_corpus/v1/csv/unbalanced_train_segments.csv",
    },
}
